import sys
sys.path.append("..")

import os
import numpy as np
import unittest
from spectralib.filterbank import create_filterbank, read_filterbank, show_filterbank
import matplotlib.pyplot as plt



class TestFilterbank(unittest.TestCase):
    
    def setUp(self):
        current_dir = os.path.dirname(os.path.abspath(__file__))
        self.input_file = os.path.join(current_dir, "ASKAP.fil")
        self.output_file = os.path.join(current_dir, "test_filterbank.fil")
        self.metadata = {'source_name': 'G348-50', 'data_type': 1, 'nchans': 336, 'barycentric': 0, 'pulsarcentric': 0, 'tsamp': 0.00126646875, 'rawdatafile': '2017-04-16-22:16:47_0000000000000000.000000.00.fil', 'src_raj': 215057.8011944, 'az_start': 0.0, 'za_start': 0.0, 'nifs': 1, 'telescope_id': 7, 'nbits': 8, 'fch1': 1488.0, 'foff': -1.0, 'src_dej': -33242.295403568, 'tstart': 57859.96643286721, 'machine_id': 0}

        # Update the shape of the data array to match the metadata
        num_samples = 10
        self.data = np.random.randint(0, 256, size=(num_samples, self.metadata['nchans']), dtype=np.uint8)

    def tearDown(self):
        if os.path.exists(self.output_file):
            os.remove(self.output_file)

    def test_read_filterbank_header(self):
        read_data, read_metadata = read_filterbank(self.input_file)
        self.assertDictEqual(self.metadata, read_metadata)

    def test_create_and_read_filterbank(self):
        create_filterbank(self.data, self.output_file,  self.metadata)
        self.assertTrue(os.path.exists(self.output_file))

        read_data, read_metadata = read_filterbank(self.output_file)
        self.assertDictEqual(self.metadata, read_metadata)

        # Reshape the read_data before comparison
        num_samples = self.data.shape[0]
        read_data_reshaped = read_data.reshape((num_samples, self.metadata['nchans']))
        np.testing.assert_array_equal(self.data, read_data_reshaped, err_msg="Read data does not match expected data")

    def test_show_filterbank(self):
        # Read the filterbank data from a file
        data, metadata = read_filterbank(self.input_file)
        
        try:
            # Attempt to display the data as a 2D image without showing it
            with plt.ioff():  # Turn off interactive mode
                show_filterbank(data)
        except Exception as e:
            self.fail(f"show_filterbank() raised an exception: {e}")


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
from spectralib.frb import generate_frb, calculate_dispersion_offsets
from spectralib.pulsar import generate_binary_pulsar, generate_solitary_pulsar, apparent_pulse_period

class TestPulsar(unittest.TestCase):

    def test_apparent_pulse_period(self):
        binary_params = {
            'rest_period': 1,
            'inclination': 0.5,
            'orbital_period': 3600,
            'start_phase': 0,
            'companion_mass': 1.0,
            'pulsar_mass': 1.4,
            'eccentricity': 0.1,
            'omega': 0.3
        }
        time = 1000
        p_apparent = apparent_pulse_period(binary_params, time)
        self.assertIsNotNone(p_apparent)

    def test_generate_binary_pulsar(self):
        data = np.zeros((64, 1024), dtype=np.uint8)
        DM = 100
        tsamp = 0.000064
        foff = -0.09765625
        fch1 = 1500
        binary_params = {
            'rest_period': 100,
            'inclination': 0.5,
            'orbital_period': 3600,
            'start_phase': 0,
            'companion_mass': 1.0,
            'pulsar_mass': 1.4,
            'eccentricity': 0.1,
            'omega': 0.3
        }
        frb_params = {
            'frb_duration': 10,
            'frb_amplitude': 50
        }

        data_with_pulsar = generate_binary_pulsar(data, DM, tsamp, foff, fch1, binary_params, **frb_params)

        self.assertIsNotNone(data_with_pulsar)
        self.assertEqual(data_with_pulsar.shape, data.shape)

    def test_generate_solitary_pulsar(self):
        data = np.zeros((64, 1024), dtype=np.uint8)
        DM = 100
        tsamp = 0.000064
        foff = -0.09765625
        fch1 = 1500
        rest_period = 100
        frb_params = {
            'frb_amplitude': 50
        }

        data_with_pulsar = generate_solitary_pulsar(data, DM, tsamp, foff, fch1, rest_period, **frb_params)

        self.assertIsNotNone(data_with_pulsar)
        self.assertEqual(data_with_pulsar.shape, data.shape)

if __name__ == '__main__':
    unittest.main()



import numpy as np

def generate_high_res_frb(data, DM, tsamp, foff, fch1, freq_upsample_factor, **frb_params):
    frb_start_index = frb_params.get('frb_start_index', 0)
    frb_duration = frb_params.get('frb_duration', 100)
    frb_amplitude = frb_params.get('frb_amplitude', 200)
    time_profile = frb_params.get('time_profile', None)
    freq_profile = frb_params.get('freq_profile', None)

    nchans, nsamp = data.shape

    if time_profile is None:
        time_profile = np.ones(frb_duration)
    if freq_profile is None:
        freq_profile = np.ones(nchans)

    # Interpolate time and frequency profiles
    time_profile_hr = np.interp(np.linspace(0, frb_duration - 1, frb_duration * freq_upsample_factor),
                                np.arange(frb_duration), time_profile)
    freq_profile_hr = np.interp(np.linspace(0, nchans - 1, nchans * freq_upsample_factor),
                                np.arange(nchans), freq_profile)

    # Calculate high-resolution dispersion offsets
    foff_hr = foff / freq_upsample_factor
    fch1_hr = np.array([fch1 + i * foff for i in range(nchans * freq_upsample_factor)])
    sub_band_offsets = calculate_dispersion_offsets(DM, fch1_hr, foff_hr, nchans * freq_upsample_factor, tsamp)

    for i in range(nchans * freq_upsample_factor):
        for j in range(frb_duration * freq_upsample_factor):
            time_index = round((frb_start_index * freq_upsample_factor) + j + sub_band_offsets[i])
            if 0 <= time_index < nsamp * freq_upsample_factor:
                data[i // freq_upsample_factor, time_index] += frb_amplitude * time_profile_hr[j] * freq_profile_hr[i]

    data = np.clip(data, 0, 255)
    return data

def generate_frb(data, DM, tsamp, foff, fch1, **frb_params):
    frb_start_index = frb_params.get('frb_start_index', 0)
    frb_duration = frb_params.get('frb_duration', 100)
    frb_amplitude = frb_params.get('frb_amplitude', 200)
    time_profile = frb_params.get('time_profile', None)
    freq_profile = frb_params.get('freq_profile', None)

    nchans, nsamp = data.shape

    if time_profile is None:
        time_profile = np.ones(frb_duration)
    if freq_profile is None:
        freq_profile = np.ones(nchans)

    offsets = calculate_dispersion_offsets(DM, fch1, foff, nchans, tsamp)

    for i in range(nchans):
        for j in range(frb_duration):
            time_index = round(frb_start_index + j + offsets[i])
            if 0 <= time_index < nsamp:
                data[i, time_index] += frb_amplitude * time_profile[j] * freq_profile[i]

    data = np.clip(data, 0, 255)
    return data

def calculate_dispersion_offsets(DM, fch1, foff, nchans, tsamp):
    freqs = np.zeros(nchans)
    for i in range(nchans):
        freqs[i] = fch1 + i * foff

    delays = np.zeros(nchans)
    for i in range(nchans):
        delays[i] = 4148.741601 * DM * ((1 / (freqs[i] ** 2)) - (1 / (freqs[0] ** 2)))

    offsets = np.zeros(nchans)
    for i in range(nchans):
        offsets[i] = round(delays[i] / tsamp)
    
    return offsets


import numpy as np
import math
import random

def generate_rfi(data, **kwargs):
    """
    Generate synthetic RFI data and save it as a filterbank file.

    :param data: The input data to add RFI to.
    :param kwargs: Additional keyword arguments.
    """
    nchans, nsamp = data.shape

    # Set default values for the keyword arguments
    default_kwargs = {
        'freqchanswidth': nchans,
        'onlength': 10,
        'ispersistent': False,
        'doesrepeat': False,
        'numrepeats': 0,
        'dutycycle': 50,  # Added default value for duty cycle
        'RFIamplitude': 200
    }

    # Update default values with provided keyword arguments
    default_kwargs.update(kwargs)

    # Unpack the updated keyword arguments
    freqchanswidth = default_kwargs['freqchanswidth']
    onlength = default_kwargs['onlength']
    ispersistent = default_kwargs['ispersistent']
    doesrepeat = default_kwargs['doesrepeat']
    numrepeats = default_kwargs['numrepeats']
    dutycycle = default_kwargs['dutycycle']  # Unpacked duty cycle
    RFIamplitude = default_kwargs['RFIamplitude']

    # Compute offlength based on the provided onlength and dutycycle
    offlength = round(onlength * (100 - dutycycle) / dutycycle)

    startfreqindex = 0
    if nchans == freqchanswidth:
        print("Adding broadband ", end="")
    else:
        print("Adding narrowband ", end="")
        startfreqindex = np.random.randint(0, nchans)

    if ispersistent and not doesrepeat:
        print("persistent RFI")
    elif not ispersistent and doesrepeat:
        print("repeating RFI, nrepeat = " + str(numrepeats))
    elif not ispersistent and not doesrepeat:
        print("impulse RFI")
    elif ispersistent and doesrepeat:
        print("persistent periodic RFI")

    endfreqindex = startfreqindex + freqchanswidth

    if ispersistent == False:
        starttimeindex = np.random.randint(0, nsamp)
        starttimeindex = starttimeindex - round(onlength/2)
        endtimeindex = starttimeindex + onlength
        if starttimeindex < 0:
            starttimeindex = 0
        if endtimeindex > nsamp:
            endtimeindex = nsamp
    else:
        starttimeindex = 0
        if doesrepeat:
            endtimeindex = starttimeindex + onlength
        else:
            endtimeindex = nsamp

    if doesrepeat:
        if ispersistent:
            while endtimeindex < nsamp:
                data[startfreqindex:endfreqindex, starttimeindex:endtimeindex] += RFIamplitude
                starttimeindex += onlength + offlength
                endtimeindex += onlength + offlength
        else:
            for i in range(-round(numrepeats/2),round(numrepeats/2),1):
                offset = i * (onlength + offlength)
                if starttimeindex + offset < 0:
                    continue
                if endtimeindex + offset > nsamp:
                    continue
                data[startfreqindex:endfreqindex, starttimeindex+offset:endtimeindex+offset] += RFIamplitude
    else:
        data[startfreqindex:endfreqindex, starttimeindex:endtimeindex] += RFIamplitude

    data = np.clip(data, 0, 255)
    return data

def add_wandering_baseline(data, **kwargs):
    # A function to add a slowly varying random signal to all channels

    default_kwargs = {
        'wanderingbaselineamplitude': 1,
        'wanderingbaselineperiod': 100,
        'custom_baseline': None
    }
    default_kwargs.update(kwargs)

    wanderingbaselineamplitude = default_kwargs['wanderingbaselineamplitude']
    wanderingbaselineperiod = default_kwargs['wanderingbaselineperiod']
    custom_baseline = default_kwargs['custom_baseline']

    nchans, nsamp = data.shape

    if custom_baseline is None:
        # Generate a random signal that is the length of the data
        wanderingbaseline = np.random.normal(0, wanderingbaselineamplitude, nsamp + wanderingbaselineperiod)

        # Lowpass filter the signal
        # This is to simulate the effect of a slowly varying baseline
        # The filter is a simple moving average
        filteredwanderingbaseline = np.zeros(nsamp)
        for i in range(nsamp):
            filteredwanderingbaseline[i] = np.mean(wanderingbaseline[i:i + wanderingbaselineperiod])
    else:
        # Use the custom provided baseline
        assert len(custom_baseline) == nsamp, "Custom baseline must have the same length as the data"
        filteredwanderingbaseline = custom_baseline


    # Add the filtered signal to each channel
    for i in range(nchans):
        data[i, :] += filteredwanderingbaseline

    data = np.clip(data, 0, 255)

    return data


def sampleloguniform(lower,upper):
    lowerlog = math.log(lower)
    upperlog = math.log(upper)
    return math.exp(random.uniform(lowerlog,upperlog))


import sys
sys.path.append("..")

import numpy as np
import unittest
from spectralib.rfi import generate_rfi, add_wandering_baseline

class TestRFI(unittest.TestCase):
    
    def setUp(self):
        nchans, nsamp = 10, 10
        self.data = np.random.normal(0,18,size=(nchans, nsamp))+127
    
    def test_generate_rfi(self):
        rfi_data = generate_rfi(self.data)
        self.assertEqual(self.data.shape, rfi_data.shape, "Data shapes don't match after RFI generation")
        self.assertNotEqual(np.sum(self.data), np.sum(rfi_data), "Data is not changed after RFI generation")

    def test_add_wandering_baseline(self):
        wanderingbaselineamplitude = 30
        wanderingbaselineperiod = 1000
        datacopy = self.data.copy()
        wb_data = add_wandering_baseline(self.data, wanderingbaselineamplitude=wanderingbaselineamplitude, wanderingbaselineperiod=wanderingbaselineperiod)
        self.assertEqual(self.data.shape, wb_data.shape, "Data shapes don't match after wandering baseline addition")
        self.assertNotEqual(np.sum(datacopy), np.sum(wb_data), "Data is not changed after wandering baseline addition")

if __name__ == '__main__':
    unittest.main()



import sys
sys.path.append("..")

import unittest
import numpy as np
from spectralib.frb import generate_frb, calculate_dispersion_offsets

class TestFRB(unittest.TestCase):
    def test_calculate_dispersion_offsets(self):
        DM = 100
        tsamp = 0.001
        foff = -0.1
        fch1 = 1000
        nchans = 512

        offsets = calculate_dispersion_offsets(DM, fch1, foff, nchans, tsamp)
        self.assertEqual(len(offsets), nchans)

    def test_generate_frb(self):
        nchans, nsamp = 100, 10000
        data = np.random.normal(0,18,size=(nchans, nsamp))+127
        DM = 100
        tsamp = 0.001
        foff = -0.5
        fch1 = 1500
        frb_params = {
            'frb_start_index': 1,
            'frb_duration': 10,
            'frb_amplitude': 50
        }
        datacopy = data.copy()
        data_with_frb = generate_frb(data, DM, tsamp, foff, fch1, **frb_params)

        self.assertIsNotNone(data_with_frb)
        self.assertEqual(data_with_frb.shape, data.shape)
        self.assertNotEqual(np.sum(datacopy), np.sum(data_with_frb))

if __name__ == '__main__':
    unittest.main()



import struct
import numpy as np
import matplotlib.pyplot as plt
import os
from spectralib.frb import calculate_dispersion_offsets

def plot_and_save(data, title, file_name):
    fig, ax = plt.subplots(figsize=(8, 6))
    im = ax.imshow(data, aspect="auto", cmap="plasma", origin="lower")
    ax.set_title(title)
    plt.colorbar(im, ax=ax)
    ax.set_ylim(ax.get_ylim()[::-1])  # Flip the y-axis
    plt.xlabel("Time (bins)")
    plt.ylabel("Frequency (bins)")
    plt.tight_layout()
    plt.savefig(file_name, dpi=300)  # Save as a high-quality PNG image
    plt.close(fig)  # Close the figure to free up memory

def dedisperse_filterbank_to_timeseries(data, DM, tsamp, foff, fch1):
    dedispersed_data = dedisperse_filterbank(data, DM, tsamp, foff, fch1)
    timeseries = np.sum(dedispersed_data, axis=0)
    return timeseries

def dedisperse_filterbank(data, DM, tsamp, foff, fch1):
    nchans, nsamp = data.shape
    dedispersed_data = np.zeros_like(data)
    offsets = calculate_dispersion_offsets(DM, fch1, foff, nchans, tsamp)

    for i in range(nchans):
        offset = int(offsets[i])
        dedispersed_data[i] = np.roll(data[i], -offset)

    return dedispersed_data


def create_filterbank(data,output_filename, metadata):
    """
    Create a filterbank file with the given data and metadata.

    :param output_filename: The output filterbank file path.
    :param data: The data to be included in the filterbank file.
    :param metadata: The metadata to be included in the filterbank file.
    """
    print(output_filename)
    with open(output_filename, 'wb') as f:
        # Write the header start marker
        # Use '<I' for little-endian unsigned int encoding
        f.write(struct.pack('<I', len('HEADER_START')))
        f.write(b'HEADER_START')
        
        # Write metadata key-value pairs
        for key, value in metadata.items():
            # Encode and write the length of the metadata key
            f.write(struct.pack('<I', len(key)))
            # Encode and write the metadata key as bytes
            f.write(key.encode())
            
            # Check the type of the metadata value and encode it accordingly
            if isinstance(value, str):
                # Encode and write the length of the string value
                f.write(struct.pack('<I', len(value)))
                # Encode and write the string value as bytes
                f.write(value.encode())
            elif isinstance(value, int):
                # Encode and write the integer value as little-endian signed int
                f.write(struct.pack('<i', value))
            elif isinstance(value, float):
                # Encode and write the float value as little-endian double precision float
                f.write(struct.pack('<d', value))
            else:
                raise ValueError(f"Unsupported data type for key '{key}': {type(value)}")
                
        # Write the header end marker
        f.write(struct.pack('<I', len('HEADER_END')))
        f.write(b'HEADER_END')
        
        # Write the signal data
        # tofile() writes the array to a binary file in a machine-specific format
        data = data.astype(np.uint8)
        data = data.transpose()
        data = np.fliplr(data)
        data.tofile(f)

def read_filterbank_data(file_path, header_params, header_len):
    nchans = header_params["nchans"]
    nbits = header_params["nbits"]

    # Validate that the data format is supported
    if nbits not in [8, 16, 32]:
        raise ValueError(f"Unsupported data format: {nbits}-bit")

    # Calculate the size of one sample in bytes
    sample_size = nbits // 8

    # Open the file and seek to the end of the header
    with open(file_path, "rb") as file:
        file.seek(header_len)

        # Read the entire file into a 1D NumPy array
        data = np.fromfile(file, dtype=np.uint8 if nbits == 8 else (np.uint16 if nbits == 16 else np.uint32))

    # Calculate the number of time samples
    nsamples = len(data) // (nchans * sample_size)

    # Reshape the data into a 2D NumPy array
    data_2d = data.reshape((nsamples, nchans))
    data_2d = data_2d.transpose()
    data_2d = np.flipud(data_2d)

    return data_2d

def read_filterbank_header(file_path):
    def get_string(file):
        nchar = struct.unpack("<i", file.read(4))[0]

        if nchar > 80 or nchar < 1:
            file.seek(-3, os.SEEK_CUR)
            string_data = file.read(4)
            print(f"Error occurred. Raw bytes: {string_data}")
            return "ERROR", 1

        string_data = file.read(nchar)
        string = string_data.decode("utf-8", "ignore")
        return string, nchar + 4

    def read_int(file):
        return struct.unpack("<i", file.read(4))[0], 4

    def read_double(file):
        return struct.unpack("<d", file.read(8))[0], 8

    header_params = {}

    with open(file_path, "rb") as file:
        param_name, nbytes = get_string(file)

        if param_name != "HEADER_START":
            file.seek(0)
            return header_params, 0
        totalbytes = nbytes

        while True:
            param_name, nbytes = get_string(file)
            totalbytes += nbytes

            if param_name == "HEADER_END":
                break

            if param_name in ["rawdatafile", "source_name"]:
                string_value, nbytes_read = get_string(file)
                header_params[param_name] = string_value
                totalbytes += nbytes_read

            elif param_name in ["az_start", "za_start", "src_raj", "src_dej", "tstart", "tsamp", "period", "fch1", "foff", "nchans", "telescope_id", "machine_id", "data_type", "ibeam", "nbeams", "nbits", "barycentric", "pulsarcentric", "nbins", "nifs", "npuls", "refdm"]:
                value, nbytes_read = read_double(file) if param_name in ["az_start", "za_start", "src_raj", "src_dej", "tstart", "tsamp", "period", "fch1", "foff"] else read_int(file)
                header_params[param_name] = value
                totalbytes += nbytes_read

            else:
                print(f"Unknown parameter: {param_name}")
                return header_params, totalbytes

    return header_params, totalbytes

def read_filterbank(file_path):
    header_params, header_len = read_filterbank_header(file_path)
    data = read_filterbank_data(file_path, header_params, header_len)
    return data, dict(header_params)


def show_filterbank(data, title='Filterbank'):
    """
    Display the filterbank data as a 2D image.

    :param data: The 2D array containing filterbank data.
    :param title: Optional title for the plot.
    """
    plt.figure()
    plt.imshow(data, aspect='auto', origin='lower', cmap='viridis')
    plt.colorbar()
    plt.title(title)
    plt.xlabel('Frequency Channels')
    plt.ylabel('Time Samples')
    plt.show()



import sys
sys.path.append("../spectralib/")
import numpy as np
import timeit
import os
from spectralib.filterbank import create_filterbank
from spectralib.pulsar import generate_binary_pulsar
from spectralib.frb import generate_frb
from spectralib.rfi import generate_rfi

def benchmark_filterbank():
    def run_create_filterbank():
        tobs = 600 # observation time in seconds
        tsamp = 0.000128 # sample time in seconds
        metadata = {
        "source_name": "SPECTRAL_FRB",
        "machine_id": 0,
        "telescope_id": 0,
        "data_type": 0,
        "fch1": 1500.0,
        "foff": -0.5,
        "nchans": 500,
        "nbits": 8,
        "tstart": 55555.0,
        "tsamp": tsamp,
        "nifs": 1,
        "nbeams": 1,
        "ibeam": 1,
        "nsamples": round(tobs/tsamp)
        }

        noisesigma = 18 #standard deviation of noise, value copied from some ASKAP data
        nchans = metadata["nchans"]
        nsamp = metadata["nsamples"]
        data = np.random.normal(0,noisesigma,size=(nchans, nsamp))+127
        output_file = "benchmark_filterbank.fil"
        create_filterbank(data,output_file, metadata)
        if os.path.exists(output_file):
            os.remove(output_file)

    return timeit.timeit(run_create_filterbank, number=10)

def benchmark_pulsar():
    def run_generate_pulsar():
        tobs = 600 # observation time in seconds
        tsamp = 0.000128 # sample time in seconds
        metadata = {
        "source_name": "SPECTRAL_FRB",
        "machine_id": 0,
        "telescope_id": 0,
        "data_type": 0,
        "fch1": 1500.0,
        "foff": -0.5,
        "nchans": 500,
        "nbits": 8,
        "tstart": 55555.0,
        "tsamp": tsamp,
        "nifs": 1,
        "nbeams": 1,
        "ibeam": 1,
        "nsamples": round(tobs/tsamp)
        }

        noisesigma = 18 #standard deviation of noise, value copied from some ASKAP data
        nchans = metadata["nchans"]
        nsamp = metadata["nsamples"]
        data = np.random.normal(0,noisesigma,size=(nchans, nsamp))+127
        DM = 10000  # Dispersion measure of the FRB

        # Binary pulsar parameters
        binary_params = {
            "inclination": np.radians(45),
            "orbital_period": 200,
            "start_phase": 0,
            "companion_mass": 1.4,
            "pulsar_mass": 1.4,
            "eccentricity": 0.1,
            "omega": np.radians(90),
        }

        # Rest pulse period of the pulsar
        p_rest = 1.0  # seconds

        # FRB parameters
        frb_params = {
            'frb_duration': 100,
            'frb_amplitude': 200,
            'time_profile': np.random.normal(1, 0.1, 100),
            'freq_profile': np.random.normal(1, 0.1, metadata['nchans']),
        }

        # Inject the binary pulsar signature
        data = generate_binary_pulsar(data, DM, metadata['tsamp'], metadata['foff'], metadata['fch1'], p_rest, binary_params, **frb_params)

        # Create the filterbank file
        output_filename = "RFIoutput_with_binary_pulsar.fil"
        create_filterbank(data,output_filename, metadata)

    return timeit.timeit(run_generate_pulsar, number=10)

def benchmark_frb():
    def run_generate_frb():
        generate_frb()

    return timeit.timeit(run_generate_frb, number=10)

def benchmark_rfi():
    def run_generate_rfi():
        generate_rfi()

    return timeit.timeit(run_generate_rfi, number=10)

if __name__ == "__main__":
    filterbank_time = benchmark_filterbank()
    pulsar_time = benchmark_pulsar()
    #frb_time = benchmark_frb()
    #rfi_time = benchmark_rfi()

    print(f"Create Filterbank: {filterbank_time:.4f} seconds")
    print(f"Generate Pulsar: {pulsar_time:.4f} seconds")
    #print(f"Generate FRB: {frb_time:.4f} seconds")
    #print(f"Generate RFI: {rfi_time:.4f} seconds")



from spectralib.frb import generate_frb, calculate_dispersion_offsets
import numpy as np

def apparent_pulse_period(binary_params, time):
    # heavily influenced by sigproc's fake, Duncan Lorimer, and Mike Keith's implementation in C
    G = 6.67430e-11
    M_SUN = 1.9885e30
    SPEED_OF_LIGHT = 299792458

    rest_period = binary_params["rest_period"]
    inclination = binary_params["inclination"]
    orbital_period = binary_params["orbital_period"]
    start_phase = binary_params["start_phase"]
    mass_companion = binary_params["companion_mass"]
    mass_pulsar = binary_params["pulsar_mass"]
    eccentricity = binary_params["eccentricity"]
    omega = binary_params["omega"]

    omega_b = 2 * np.pi / orbital_period
    t0 = start_phase * orbital_period
    mass_function = (mass_companion * np.sin(inclination))**3 / (mass_companion + mass_pulsar)**2
    asini = (G * M_SUN * mass_function * orbital_period**2 / (4 * np.pi**2))**(1/3)

    mean_anomaly = omega_b * (time - t0)
    eccentric_anomaly = mean_anomaly

    for _ in range(10):
        e_next = eccentric_anomaly - (eccentric_anomaly - eccentricity * np.sin(eccentric_anomaly) - mean_anomaly) / (1 - eccentricity * np.cos(eccentric_anomaly))
        if abs(e_next - eccentric_anomaly) < 1e-10:
            break
        eccentric_anomaly = e_next

    true_anomaly = 2 * np.arctan(np.sqrt((1 + eccentricity) / (1 - eccentricity)) * np.tan(eccentric_anomaly / 2))
    velocity = omega_b * asini / np.sqrt(1 - eccentricity**2) * (np.cos(omega + true_anomaly) + eccentricity * np.cos(omega))
    p_apparent = rest_period * (1 + velocity / SPEED_OF_LIGHT)

    return p_apparent

def generate_binary_pulsar(data, DM, tsamp, foff, fch1, binary_params, **frb_params):
    nchans, nsamp = data.shape
    
    # Use calculate_dispersion_offsets from frb.py to calculate the maximum offset, and subtract it from 0 to get the start index
    # this ensures the pulsar signal is present in the entire data array
    offsets = calculate_dispersion_offsets(DM, fch1, foff, nchans, tsamp)
    start_index = -max(offsets)

    pulse = 0
    pulse_start_time = start_index*tsamp
    while pulse_start_time < nsamp*tsamp:
        pulse_start_time = start_index*tsamp + pulse * binary_params["rest_period"]
        app_pulse_period = apparent_pulse_period(binary_params, pulse_start_time)
        print("At time : ",pulse_start_time," apparent pulse period: ", app_pulse_period)
        frb_start_time = pulse_start_time + app_pulse_period
        frb_start_index = int(frb_start_time / tsamp)

        data = generate_frb(data, DM, tsamp, foff, fch1, frb_start_index=frb_start_index, **frb_params)
        pulse+=1

    return data

def generate_solitary_pulsar(data, DM, tsamp, foff, fch1, rest_period, **frb_params):
    nchans, nsamp = data.shape
    pulse_duration = frb_params.get("frb_duration", 100)


    # Use calculate_dispersion_offsets from frb.py to calculate the maximum offset, and subtract it from 0 to get the start index
    # this ensures the pulsar signal is present in the entire data array, rather than missing from the bottom left corner
    offsets = calculate_dispersion_offsets(DM, fch1, foff, nchans, tsamp)
    start_index = -max(offsets)
    pulse = 0
    pulse_start_time = start_index*tsamp
    while pulse_start_time < nsamp*tsamp:
        pulse_start_time = start_index*tsamp + pulse * rest_period
        frb_start_index = int(pulse_start_time / tsamp)

        data = generate_frb(data, DM, tsamp, foff, fch1, frb_start_index=frb_start_index, frb_duration=pulse_duration, **frb_params)
        pulse+=1

    return data


from setuptools import setup, find_packages

setup(
    name="spectralib",
    version="0.0.11",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    classifiers=[
        "Development Status :: 1 - Planning",
    ],
    author="Jack White",
    author_email="jack.white@eng.ox.ac.uk",
    description="Synthetic Pulsar Emission, Contamination and Transients Radio Astronomy Library (SPECTRALib)",
    long_description="A python library for creating synthetic filterbank files (pulsars, FRBs, RFI) for radio astronomy. Integrate functions in your code or use standalone to generate custom datasets.",
    long_description_content_type="text/markdown",
    url="https://github.com/jack-white1/spectralib",
)
